/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.xml.transform.Source;

import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;

/**
 * Change the MEP of an exchange. Use the MEP defined in the Consumes element. <br>
 * Depending on the MEP, some response can be lost.
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * 
 */
public class Bridge extends AbstractPattern {

    public final static String BRIDGE_NAMESPACE = EIPConstants.NAMESPACE + "/bridge";

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals.
     * component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final Exchange originalExchange = exchange;
            final List<Consumes> consumesList = context.getSUConsumes(originalExchange
                    .getEndpoint());

            if (consumesList != null && consumesList.size() == 1) {
                final Consumes consumes = consumesList.get(0);
                // create the exchange with the new MEP to send to the next
                // service
                Exchange newExchange;
                newExchange = context.createConsumeExchange(consumes);
                PatternHelper.copy(originalExchange.getInMessage(), newExchange.getInMessage());
                UtilFactory.getExchangeUtil().copyExchangeProperties(originalExchange, newExchange);
                // send the exchange to the next service
                this.logger.fine(this.getName() + ": original M.E.P. : "
                        + originalExchange.getPattern());
                this.logger
                .fine(this.getName() + ": target   M.E.P. : " + newExchange.getPattern());
                this.logSend(consumes, newExchange);
                context.sendSync(newExchange);
                // process the MEP mapping
                this.processMEPMapping(originalExchange, newExchange, context);
                // close the forwarded exchange if active
                if (newExchange.isActiveStatus()) {
                    this.closeForwardedExchange(originalExchange, newExchange, context);
                }
            } else {
                exchange
                .setError(new JBIException(
                        this.getName()
                        + ": the Service Unit must define an unique Consumes sections for the pattern"));
            }
        } catch (final MessagingException e) {
            exchange.setError(new Exception(e));
        } catch (final PEtALSCDKException e) {
            exchange.setError(new Exception(e));
        }
        return true;
    }

    /**
     * Process the mapping between the original and new MEP : <li>If an error
     * has been received, set it on the original MEP.</li> <li>Depending on the
     * original MEP, copy received Fault or Out message to the original
     * Exchange, if any; otherwise copy an empty Out message.</li>
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    protected void processMEPMapping(final Exchange originalExchange, final Exchange newExchange,
            final ExchangeContext context) throws MessagingException, PEtALSCDKException {

        this.logger.fine(this.getName() + ": process the response");
        // copy back the error from the forwarded exchange if any
        if (newExchange.isErrorStatus()) {
            this.setErrorOnOriginalExchange(originalExchange, newExchange, context);
        } else {
            // MEP mapping
            if (originalExchange.isInOnlyPattern()) {
                // convert Fault to Error if any ...
                if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                    this.logger.fine(this.getName() + ": send back the fault as an Error");
                    final Source faultMsg = newExchange.getFault().getContent();
                    final String message = UtilFactory.getSourceUtil().createString(faultMsg);
                    final Exception exception = new Exception(message);
                    originalExchange.setError(exception);
                } else {
                    this.logger.fine(this.getName() + ": send back nothing");
                }

            } else if (originalExchange.isRobustInOnlyPattern()) {
                // send back Fault if any
                if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                    this.logger.fine(this.getName() + ": send back the fault");
                    this.setFaultOnOriginalExchange(originalExchange, newExchange);
                } else {
                    this.logger.fine(this.getName() + ": send back nothing");
                }
            } else if (originalExchange.isInOptionalOutPattern()) {
                // send back Fault or Out if any
                if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                    this.logger.fine(this.getName() + ": send back the fault");
                    this.setFaultOnOriginalExchange(originalExchange, newExchange);
                } else if (newExchange.isActiveStatus() && newExchange.getOutMessage() != null) {
                    this.logger.fine(this.getName() + ": send back the response");
                    this.setOutOnOriginalExchange(originalExchange, newExchange);
                } else {
                    this.logger.fine(this.getName() + ": send back nothing");
                }
            } else if (originalExchange.isInOutPattern()) {
                // send back Fault or Out if any, otherwise return an empty Out
                if (newExchange.isActiveStatus() && newExchange.getFault() != null) {
                    this.logger.fine(this.getName() + ": send back the fault");
                    this.setFaultOnOriginalExchange(originalExchange, newExchange);
                } else if (newExchange.isActiveStatus() && newExchange.getOutMessage() != null) {
                    this.logger.fine(this.getName() + ": send back the response");
                    this.setOutOnOriginalExchange(originalExchange, newExchange);
                } else if (newExchange.isDoneStatus()) {
                    this.logger.fine(this.getName() + ": send back a standard response");
                    this.setEmptyOutOnOriginalExchange(originalExchange);
                }
            }
        }
    }

    /**
     * Set the original exchange to the Error of the forwarded exchange
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     */
    protected void setErrorOnOriginalExchange(final Exchange originalExchange,
            final Exchange newExchange, final ExchangeContext context) {
        context
        .getLogger()
        .finest(
                this.getName()
                + " - set on the original exchange an error received from the forwarded exchange");
        originalExchange.setError(newExchange.getError());
    }

    /**
     * copy on the original exchange the Fault of the forwarded exchange If the
     * fault can not be copied or set, set an error on the original exchange
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     * @throws PEtALSCDKException
     * @throws MessagingException
     */
    protected void setOutOnOriginalExchange(final Exchange originalExchange,
            final Exchange newExchange) throws MessagingException, PEtALSCDKException {
        this.logger
        .fine(this.getName()
                + ": set on the original exchange an Out message received from the forwarded exchange");
        PatternHelper.copy(newExchange.getOutMessage(), originalExchange.getOutMessage());
    }

    /**
     * set a default message on the original exchange. set an error on the
     * original exchange if message can not be set
     * 
     * @param originalExchange
     * @param context
     * @throws MessagingException
     */
    protected void setEmptyOutOnOriginalExchange(final Exchange originalExchange)
    throws MessagingException {
        final Document document = PatternHelper.createResultDocument(BRIDGE_NAMESPACE);
        originalExchange.setOutMessageContent(document);
    }

    /**
     * copy on the original exchange the Fault of the forwarded exchange If the
     * fault can not be copied or set, set an error on the original exchange
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     * 
     * @throws MessagingException
     * @throws PEtALSCDKException
     */
    protected void setFaultOnOriginalExchange(final Exchange originalExchange,
            final Exchange newExchange) throws MessagingException, PEtALSCDKException {
        this.logger.fine(this.getName()
                + ": set on the original exchange a Fault received from the forwarded exchange");
        originalExchange.setFault(newExchange.getFault());
    }

    /**
     * Close the forwarded exchange by sending a DONE status. If the status can
     * not be sent, set an error on the original exchange
     * 
     * @param originalExchange
     * @param newExchange
     * @param context
     * @throws MessagingException
     */
    protected void closeForwardedExchange(final Exchange originalExchange,
            final Exchange newExchange, final ExchangeContext context) throws MessagingException {
        this.logger.fine(this.getName() + ": close the forwarded exchange");
        newExchange.setDoneStatus();
        context.send(newExchange);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Bridge";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        exchange.setError(new Exception(this.getName()
                + ": the pattern doesn't process the asynchronous exchanges"));
        return true;
    }

}
