/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;

import javax.jbi.JBIException;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.stream.InputStreamForker;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;

/**
 * 
 * Dispatch the exchange to all Consumes services defined in the service unit. <br>
 * <br>
 * - Accept InOnly
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * @author Frederic Gardes
 */
public class Dispatcher extends AbstractPattern {

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.Pattern#processPattern(org.ow2.petals.
     * component.framework.util.Exchange, org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final Exchange originalExchange = exchange;
            final List<Consumes> consumesList = context.getSUConsumes(originalExchange
                    .getEndpoint());

            if (consumesList != null && consumesList.size() != 0) {
                this.sendAsyncExchanges(consumesList, exchange, context);
            } else {
                exchange
                .setError(new Exception(
                        this.getName()
                        + ": the Service Unit must define at least 1 Consumes sections for the pattern "));
            }
        } catch (final PEtALSCDKException e) {
            exchange.setError(new Exception(e));
        } catch (final MessagingException e) {
            exchange.setError(new Exception(e));
        }
        return true;
    }

    /**
     * Check InOnly or RobustInOnly pattern. <br>
     * 
     * @param uri
     * @return true if MEP is InOnly, false otherwise
     */
    @Override
    public boolean validateMEP(final URI uri) {
        return MEPConstants.IN_ONLY_PATTERN.equals(uri);
    }

    /**
     * Send a copy of the exchange's IN message to all the services to consume.
     * 
     * @param consumesList
     * @param exchange
     * @param context
     * @throws JBIException
     * @throws PEtALSCDKException
     */
    protected void sendAsyncExchanges(final List<Consumes> consumesList, final Exchange exchange,
            final ExchangeContext context) throws PEtALSCDKException, MessagingException {

        final NormalizedMessage inMessage = exchange.getInMessage();

        // Handle the Streaming issue
        InputStreamForker inMessageAsStream = PatternHelper.handleStreamSource(inMessage);

        // send all exchanges asynchronously
        for (final Consumes toConsume : consumesList) {
            if (inMessageAsStream != null) {
                // Handle the Streaming issue
                final Source inSource = new StreamSource(inMessageAsStream.getInputStreamOne());
                inMessageAsStream = new InputStreamForker(inMessageAsStream.getInputStreamTwo());
                inMessage.setContent(inSource);
            }
            final Exchange dipatchedExchange = context.createConsumeExchange(toConsume);
            PatternHelper.copy(inMessage, dipatchedExchange.getInMessage());
            UtilFactory.getExchangeUtil().copyExchangeProperties(exchange, dipatchedExchange);
            if (dipatchedExchange.getOperation() == null) {
                dipatchedExchange.setOperation(exchange.getOperation());
            }
            this.logSend(toConsume, dipatchedExchange);
            context.send(dipatchedExchange);
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Dispatcher";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        exchange.setError(new Exception(this.getName()
                + ": the pattern doesn't process the asynchronous exchanges"));
        return true;
    }
}
