/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.ArrayList;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.ExchangeStatus;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;

import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.AbstractAsyncContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.ow2.petals.se.eip.async.RoutingSlipAsyncContext;

/**
 * Pipe all Consumes services defined in the service unit. If a Fault or Error
 * occurs on a call, stop the process and return the Fault or Error
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * @author Frederic Gardes
 * 
 */
public class RoutingSlip extends AbstractPattern {

    public final static String ROUTING_SLIP_NAME = "Routing-Slip";

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        boolean reply = false;
        try {
            final List<Consumes> consumesList = new ArrayList<Consumes>(context
                    .getSUConsumes(exchange.getEndpoint()));
            if (consumesList != null && consumesList.size() >= 1) {
                // create the exchange to pipe to the next service
                final Consumes consumes = consumesList.get(0);
                final Exchange pipedExchange = context.createConsumeExchange(consumes);
                PatternHelper.copy(exchange.getInMessage(), pipedExchange.getInMessage());
                UtilFactory.getExchangeUtil().copyExchangeProperties(exchange, pipedExchange);
                // send the exchange to the next service
                this.logSend(consumes, pipedExchange);
                context.sendAsync(pipedExchange, new RoutingSlipAsyncContext(exchange,
                        AbstractAsyncContext.DEFAULT_TIME_TO_LIVE, consumesList));
            } else {
                exchange
                .setError(new JBIException(
                        this.getName()
                        + ": the Service Unit must define at least 1 Consumes sections for the pattern "));
                reply = true;
            }
        } catch (final MessagingException e) {
            exchange.setError(new Exception(e));
            reply = true;
        } catch (final PEtALSCDKException e) {
            exchange.setError(new Exception(e));
            reply = true;
        }
        return reply;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        final boolean reply = false;
        NormalizedMessage receivedMessage = null;
        if (this.logger == null) {
            this.logger = context.getLogger();
            if (this.logger == null) {
                this.logger = Logger.getLogger(AbstractPattern.class.getName());
            }
        }
        try {
            if (asyncContext instanceof RoutingSlipAsyncContext) {
                // process the response
                receivedMessage = PatternHelper.processSentExchange(exchange, context);
                final int subMessages = asyncContext.decrementAndGetOngoingSubMessages();
                if (subMessages == 0) {
                    // For the last call, use the original M.E.P.
                    this.logger.fine(this.getName() + ": terminate the pattern");
                    // if the last called service return an out response,
                    // send it back to the original exchange.
                    // otherwise, the' processSentExchange()' has set
                    // information on the original exchange
                    if (receivedMessage != null) {
                        if (receivedMessage instanceof Fault) {
                            if (!asyncContext.getOriginalExchange().isInOnlyPattern()) {
                                asyncContext.getOriginalExchange()
                                .setFault((Fault) receivedMessage);
                            }
                        } else {
                            if (!asyncContext.getOriginalExchange().isInOnlyPattern()
                                    && !asyncContext.getOriginalExchange().isRobustInOnlyPattern()) {
                                asyncContext.getOriginalExchange().setOutMessage(receivedMessage);
                            }
                        }
                    } else if (asyncContext.getOriginalExchange().getFault() == null
                            && ExchangeStatus.ACTIVE.equals(asyncContext.getOriginalExchange()
                                    .getStatus())) {
                        asyncContext.getOriginalExchange().setDoneStatus();
                    }
                    // terminate the original exchange
                    context.send(asyncContext.getOriginalExchange());
                } else if (receivedMessage != null) {
                    // create the exchange to pipe to the next service
                    final Consumes consumes = ((RoutingSlipAsyncContext) asyncContext)
                    .getConsumesList().get(
                            ((RoutingSlipAsyncContext) asyncContext).getConsumesList()
                            .size()
                            - subMessages);
                    final Exchange pipedExchange = context.createConsumeExchange(consumes);
                    if (receivedMessage instanceof Fault) {
                        if (!asyncContext.getOriginalExchange().isInOnlyPattern()) {
                            asyncContext.getOriginalExchange().setFault((Fault) receivedMessage);
                        }
                        // terminate the original exchange
                        context.send(asyncContext.getOriginalExchange());
                    } else {
                        PatternHelper.copy(receivedMessage, pipedExchange.getInMessage());
                        UtilFactory.getExchangeUtil().copyExchangeProperties(
                                asyncContext.getOriginalExchange(), pipedExchange);
                        // send the exchange to the next service
                        this.logSend(consumes, pipedExchange);
                        context.sendAsync(pipedExchange, asyncContext);
                    }
                } else {
                    if (asyncContext.getOriginalExchange().isActiveStatus()) {
                        asyncContext
                        .getOriginalExchange()
                        .setError(
                                new Exception(
                                "Terminate the pattern processing due to a Done status received before the end of the global routing"));
                        context.send(asyncContext.getOriginalExchange());
                    }
                }
            } else {
                if (asyncContext.getOriginalExchange().isActiveStatus()) {
                    asyncContext
                    .getOriginalExchange()
                    .setError(
                            new Exception(
                            "The asynchronous context isn't a routing slip context: unable to process"));
                }
            }
        } catch (final JBIException e) {
            if (asyncContext.getOriginalExchange().isActiveStatus()) {
                asyncContext.getOriginalExchange().setError(new Exception(e));
            }
            // in this case, it's an error, terminate the original exchange
            try {
                context.send(asyncContext.getOriginalExchange());
            } catch (final MessagingException e1) {
                this.logger.warning(this.getName()
                        + ": unable to close the source exchange on error status");
            }
        }
        return reply;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return ROUTING_SLIP_NAME;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return true;
    }

}
