/**
 * PETALS - PETALS Services Platform. Copyright (c) 2007 EBM Websourcing,
 * http://www.ebmwebsourcing.com/
 * 
 * This library is free software; you can redistribute it and/or modify it under
 * the terms of the GNU Lesser General Public License as published by the Free
 * Software Foundation; either version 2.1 of the License, or (at your option)
 * any later version. This library is distributed in the hope that it will be
 * useful, but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU Lesser
 * General Public License for more details.
 * 
 * You should have received a copy of the GNU Lesser General Public License
 * along with this library; if not, write to the Free Software Foundation, Inc.,
 * 59 Temple Place, Suite 330, Boston, MA 02111-1307 USA
 * 
 * -------------------------------------------------------------------------
 * $Id$
 * -------------------------------------------------------------------------
 */

package org.ow2.petals.se.eip.patterns;

import java.net.URI;
import java.util.List;
import java.util.logging.Logger;

import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;

import org.ow2.petals.commons.stream.InputStreamForker;
import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.AbstractAsyncContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;

/**
 * Aggregate all Consumes services defined in the service unit. <br>
 * Return the aggregated results.<br>
 * <br>
 * - Accept InOut<br>
 * - If a Fault or Error occurs on a call, stop the process and return the Fault
 * or Error Note : The exchange will be sent back automatically by the framework
 * 
 * @author alouis, chamerling - eBM WebSourcing
 * @author Frederic Gardes
 * 
 */
public class ScatterGather extends AbstractPattern {

    public final static String SCATTER_GATHER_NAMESPACE = EIPConstants.NAMESPACE
    + "/scatter-gather";

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.util.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        boolean reply = false;
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());
            if (consumesList != null && consumesList.size() != 0) {
                this.sendAsyncExchanges(consumesList, exchange, context);
            } else {
                exchange
                .setError(new Exception(
                        this.getName()
                        + ": the Service Unit must define at least 1 Consumes sections for the pattern"));
                reply = true;
            }
        } catch (final JBIException e) {
            exchange.setError(new Exception(e));
            reply = true;
        }
        return reply;
    }

    /**
     * Send a copy of the exchange's IN message to all the services to consume.
     * Return the list of the Exchanges sent asynchronously. If an error occurs,
     * set a Fault on the original exchange, and return null
     * 
     * @param consumesList
     * @param exchange
     * @param context
     * @return the list of the exchanges sent async, or null if a problem
     *         occurred
     * @throws JBIException
     */
    private void sendAsyncExchanges(final List<Consumes> consumesList, final Exchange exchange,
            final ExchangeContext context) throws JBIException {
        Document document = null;
        InputStreamForker inMessageAsStream = null;
        boolean faultRobust = false;
        boolean exceptionRobust = false;
        if (context.getExtensions() != null) {
            final String stringValue = context.getExtensions().get(EIPConstants.FAULT_ROBUST);
            if (stringValue != null) {
                faultRobust = Boolean.valueOf(stringValue);
            }
        }
        if (context.getExtensions() != null) {
            final String stringValue = context.getExtensions().get(EIPConstants.EXCEPTION_ROBUST);
            if (stringValue != null) {
                exceptionRobust = Boolean.valueOf(stringValue);
            }
        }
        final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                AbstractAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(), faultRobust,
                exceptionRobust, consumesList.size());
        final NormalizedMessage inMessage = exchange.getInMessage();
        // Handle the Streaming issue
        inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
        if (inMessageAsStream == null) {
            if (inMessage.getContent() instanceof DOMSource) {
                // Handle the DOM issue
                document = UtilFactory.getSourceUtil().createDocument(inMessage.getContent());
            } else {
                throw new JBIException("The content must be a dom source or a stream source");
            }
        }
        // send all exchanges asynchronously
        for (final Consumes consumedService : consumesList) {
            final Exchange tmpExchange = context.createConsumeExchange(consumedService);
            PatternHelper.copy(inMessage, tmpExchange.getInMessage());
            UtilFactory.getExchangeUtil().copyExchangeProperties(exchange, tmpExchange);
            this.logSend(consumedService, tmpExchange);
            context.sendAsync(tmpExchange, asyncContext);
            if (inMessageAsStream != null) {
                // Handle the Streaming issue
                inMessage.setContent(new StreamSource(inMessageAsStream.getInputStreamTwo()));
                inMessageAsStream = PatternHelper.handleStreamSource(inMessage);
            } else {
                // Handle the DOM issue
                inMessage.setContent(UtilFactory.getSourceUtil().createDOMSource(document));
            }
        }
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        NormalizedMessage receivedMessage = null;
        boolean finished = false;

        if (this.logger == null) {
            this.logger = context.getLogger();
            if (this.logger == null) {
                this.logger = Logger.getLogger(AbstractPattern.class.getName());
            }
        }
        synchronized (asyncContext.getOriginalExchange()) {
            try {
                final int subMessages = asyncContext.decrementAndGetOngoingSubMessages();
                if (subMessages == 0) {
                    finished = true;
                }
                if (asyncContext.getOriginalExchange().isActiveStatus()) {
                    receivedMessage = PatternHelper.processSentExchange(exchange, context);
                    if (receivedMessage != null) {
                        if (receivedMessage instanceof Fault) {
                            // fault
                            if (!asyncContext.isFaultRobust()) {
                                if (asyncContext.getOriginalExchange().getFault() == null
                                        && asyncContext.getOriginalExchange().getError() == null) {
                                    asyncContext.getOriginalExchange().setFault(
                                            (Fault) receivedMessage);
                                    finished = true;
                                }
                            } else {
                                // robustness on an external exchange
                                this.aggregate(asyncContext, receivedMessage);
                            }
                        } else {
                            // out
                            this.aggregate(asyncContext, receivedMessage);
                        }
                    } else {
                        // Done
                        Document document = this.getAggregatedDocument(asyncContext);
                        if (document == null) {
                            document = PatternHelper.createResultDocument(SCATTER_GATHER_NAMESPACE);
                        }
                        asyncContext.getOriginalExchange().getOutMessage().setContent(
                                UtilFactory.getSourceUtil().createStreamSource(document));
                    }
                }
            } catch (final JBIException e) {
                // Error
                if (asyncContext.getOriginalExchange().isActiveStatus()
                        && asyncContext.getOriginalExchange().getError() == null) {
                    if (!asyncContext.isExceptionRobust()) {
                        asyncContext.getOriginalExchange().setError(new Exception(e));
                        finished = true;
                    } else {
                        // robustness on an external exchange
                        try {
                            this.aggregate(asyncContext, e);
                        } catch (final JBIException e2) {
                            asyncContext.getOriginalExchange().setError(new Exception(e2));
                            finished = true;
                        }
                    }
                }
            }
            if (finished) {
                this.logger.fine(this.getName() + ": terminate the pattern");
                if (!asyncContext.isTerminated()) {
                    try {
                        if (asyncContext.getOriginalExchange().isActiveStatus()
                                && asyncContext.getOriginalExchange().getFault() == null
                                && asyncContext.getOriginalExchange().getError() == null) {
                            final Document document = UtilFactory.getSourceUtil()
                            .createDocument(
                                    asyncContext.getOriginalExchange().getOutMessage()
                                    .getContent());
                            document.normalizeDocument();
                            asyncContext.getOriginalExchange().getOutMessage().setContent(
                                    UtilFactory.getSourceUtil().createStreamSource(document));
                        }
                        context.send(asyncContext.getOriginalExchange());
                    } catch (final MessagingException e) {
                        this.logger.warning(this.getName() + ": unable to send an exchange, "
                                + e.getMessage());
                    } catch (final PEtALSCDKException e) {
                        this.logger.warning(this.getName() + ": unable to send an exchange, "
                                + e.getMessage());
                    }
                    asyncContext.setTerminated();
                }
            }
        }
        return true;
    }

    /**
     * Aggregate the string content of the exception to the source exchange
     * 
     * @param asyncContext
     *            The context where the source exchange is stored
     * @param exception
     *            The received exception
     * @throws JBIException
     */
    private void aggregate(final AbstractAsyncContext asyncContext, final Exception exception)
    throws JBIException {
        // get the document from the source exchange
        Document aggregatedDocument = this.getAggregatedDocument(asyncContext);
        final Document itemDocument = PatternHelper
        .createExceptionDocument(SCATTER_GATHER_NAMESPACE);
        if (itemDocument != null) {
            itemDocument.getDocumentElement().setTextContent(exception.getMessage());
            aggregatedDocument = PatternHelper.aggregateContent(itemDocument, aggregatedDocument,
                    EIPConstants.RESULT_ELEMENT, SCATTER_GATHER_NAMESPACE);
            asyncContext.getOriginalExchange().getOutMessage().setContent(
                    UtilFactory.getSourceUtil().createStreamSource(aggregatedDocument));
        }
    }

    /**
     * Aggregate the XML content, attachments and properties at the content of
     * the out normalized message to the source exchange
     * 
     * @param asyncContext
     *            The context where the source exchange is stored
     * @param receivedMessage
     *            The received normalized message
     * @throws JBIException
     */
    private void aggregate(final AbstractAsyncContext asyncContext,
            final NormalizedMessage receivedMessage) throws JBIException {
        // get the document from the source exchange
        Document aggregatedDocument = this.getAggregatedDocument(asyncContext);
        aggregatedDocument = PatternHelper.aggregateContent(receivedMessage, aggregatedDocument,
                EIPConstants.RESULT_ELEMENT, SCATTER_GATHER_NAMESPACE);
        asyncContext.getOriginalExchange().getOutMessage().setContent(
                UtilFactory.getSourceUtil().createStreamSource(aggregatedDocument));
        PatternHelper.copyAttachments(receivedMessage, asyncContext.getOriginalExchange()
                .getOutMessage());
        PatternHelper.copyProperties(receivedMessage, asyncContext.getOriginalExchange()
                .getOutMessage());
    }

    /**
     * Get the aggregated document into the source exchange
     * 
     * @param asyncContext
     *            The context with the source exchange
     * @return The document build from the content of the out normalized message
     * @throws PEtALSCDKException
     */
    private Document getAggregatedDocument(final AbstractAsyncContext asyncContext)
    throws PEtALSCDKException {
        Document aggregatedDocument = null;
        this.logger.fine(this.getName() + ": aggregates a document to the original out message");
        final NormalizedMessage normalizedMessage = asyncContext.getOriginalExchange()
        .getOutMessage();
        if (normalizedMessage != null) {
            final Source outContent = normalizedMessage.getContent();
            if (outContent != null) {
                aggregatedDocument = UtilFactory.getSourceUtil().createDocument(outContent);
            }
        }
        return aggregatedDocument;
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Scatter-Gather";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return MEPConstants.IN_OUT_PATTERN.equals(mep);
    }
}
