/**
 * The Splitter EIP pattern
 */

package org.ow2.petals.se.eip.patterns;

import java.io.IOException;
import java.net.URI;
import java.util.ArrayList;
import java.util.List;

import javax.activation.DataHandler;
import javax.activation.DataSource;
import javax.jbi.JBIException;
import javax.jbi.messaging.Fault;
import javax.jbi.messaging.MessagingException;
import javax.jbi.messaging.NormalizedMessage;
import javax.xml.transform.Source;
import javax.xml.transform.dom.DOMSource;
import javax.xml.transform.stream.StreamSource;
import javax.xml.xpath.XPathExpressionException;

import org.ow2.petals.component.framework.api.Message.MEPConstants;
import org.ow2.petals.component.framework.api.exception.PEtALSCDKException;
import org.ow2.petals.component.framework.api.message.Exchange;
import org.ow2.petals.component.framework.jbidescriptor.generated.Consumes;
import org.ow2.petals.component.framework.util.StringHelper;
import org.ow2.petals.component.framework.util.UtilFactory;
import org.ow2.petals.se.eip.EIPConstants;
import org.ow2.petals.se.eip.ExchangeContext;
import org.ow2.petals.se.eip.async.AbstractAsyncContext;
import org.ow2.petals.se.eip.async.CommonAsyncContext;
import org.w3c.dom.Document;
import org.w3c.dom.NodeList;

/**
 * @author Adrien LOUIS - EBMWebSourcing
 * @author Frederic GARDES
 * 
 */
/**
 * @author fgardes
 * 
 */
public class Splitter extends AbstractPattern {

    public final static String SPLITTER_NAMESPACE = EIPConstants.NAMESPACE + "/splitter";

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#getName()
     */
    public String getName() {
        return "Splitter";
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#process(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext)
     */
    @Override
    public boolean process(final Exchange exchange, final ExchangeContext context) {
        try {
            final List<Consumes> consumesList = context.getSUConsumes(exchange.getEndpoint());

            if (consumesList != null && consumesList.size() == 1) {
                final Consumes consumes = consumesList.get(0);
                List<Source> splittedSources = null;
                if (!this.isAttachmentMode(context.getExtensions())) {
                    // content mode
                    final String splitterPath = context.getExtensions().get(EIPConstants.PATH);
                    if (!StringHelper.isNullOrEmpty(splitterPath)) {
                        this.logger.fine(this.getName() + ": splitting XPath : " + splitterPath);
                        final Document inMessageContent = exchange
                        .getInMessageContentAsDocument(true);
                        final NodeList nodeList = PatternHelper.evaluateXPathExpressionAsNodeList(
                                splitterPath, inMessageContent);
                        if (nodeList != null && nodeList.getLength() != 0) {
                            this.logger.fine(this.getName() + ": splitted elements : "
                                    + nodeList.getLength());

                            splittedSources = new ArrayList<Source>(nodeList.getLength());
                            for (int i = 0; i < nodeList.getLength(); i++) {
                                splittedSources.add(new DOMSource(nodeList.item(i)));
                            }
                        } else {
                            exchange.setError(new JBIException(this.getName()
                                    + ": no splitted element has been matched!"));
                        }
                    } else {
                        exchange.setError(new JBIException(this.getName()
                                + ": the Service Unit must define a 'path' element"));
                    }
                } else {
                    // attachement mode
                    splittedSources = new ArrayList<Source>();
                    if (exchange.getInMessageAttachments().size() == 0) {
                        for (final DataHandler dataHandler : exchange.getInMessageAttachments()) {
                            final String name = dataHandler.getName();
                            final String type = dataHandler.getContentType();
                            this.logger.fine(this.getName() + ": type of the current attachment "
                                    + name + ": " + type);
                            if ("application/xml".equals(type)) {
                                final DataSource content = dataHandler.getDataSource();
                                splittedSources.add(new StreamSource(content.getInputStream()));
                            }
                        }
                    } else {
                        exchange.setError(new JBIException(this.getName()
                                + ": no splitted element has been matched!"));
                    }
                }
                // check if no error was set
                if (splittedSources != null && exchange.isActiveStatus()) {
                    final boolean faultRobust = Boolean.valueOf(context.getExtensions().get(
                            EIPConstants.FAULT_ROBUST));
                    final boolean errorRobust = Boolean.valueOf(context.getExtensions().get(
                            EIPConstants.EXCEPTION_ROBUST));
                    final CommonAsyncContext asyncContext = new CommonAsyncContext(exchange,
                            AbstractAsyncContext.DEFAULT_TIME_TO_LIVE, this.getName(), faultRobust,
                            errorRobust, splittedSources.size());

                    for (final Source splittedSource : splittedSources) {
                        final Exchange splittedExchange = context.createConsumeExchange(consumes);
                        if (splittedExchange.getOperation() == null) {
                            splittedExchange.setOperation(exchange.getOperation());
                        }
                        PatternHelper
                        .copy(exchange.getInMessage(), splittedExchange.getInMessage());
                        splittedExchange.getInMessage().setContent(splittedSource);
                        UtilFactory.getExchangeUtil().copyExchangeProperties(exchange,
                                splittedExchange);
                        this.logSend(consumes, splittedExchange);
                        context.sendAsync(splittedExchange, asyncContext);
                    }
                }
            } else {
                exchange.setError(new JBIException(this.getName()
                        + ": the Service Unit must define only one consumes section"));
            }
        } catch (final MessagingException e) {
            exchange.setError(new Exception(e));
        } catch (final XPathExpressionException e) {
            exchange.setError(new Exception(e));
        } catch (final IOException e) {
            exchange.setError(new Exception(e));
        } catch (final PEtALSCDKException e) {
            exchange.setError(new Exception(e));
        }
        return true;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#processAsync(org.ow2.petals
     * .component.framework.api.message.Exchange,
     * org.ow2.petals.se.eip.ExchangeContext,
     * org.ow2.petals.se.eip.async.CommonAsyncContext)
     */
    @Override
    public boolean processAsync(final Exchange exchange, final ExchangeContext context,
            final CommonAsyncContext asyncContext) {
        NormalizedMessage receivedMessage = null;
        Document aggregatedDocument = null;
        synchronized (asyncContext.getOriginalExchange()) {
            try {
                receivedMessage = PatternHelper.processSentExchange(exchange, context);
                if (!asyncContext.isTerminated()) {
                    if (receivedMessage != null) {
                        // it can be only a fault, pattern RobustInOnly
                        if (asyncContext.isFaultRobust()) {
                            // aggregate the XML content, attachments and
                            // properties
                            this.logger.fine(this.getName()
                                    + ": aggregate a document to the FAULT message");
                            Fault aggregatedFault = asyncContext.getOriginalExchange().getFault();
                            if (aggregatedFault == null) {
                                aggregatedFault = asyncContext.getOriginalExchange().createFault();
                                aggregatedFault.setContent(new DOMSource(PatternHelper
                                        .createResultDocument(SPLITTER_NAMESPACE)));
                                asyncContext.getOriginalExchange().setFault(aggregatedFault);
                            }
                            aggregatedDocument = (Document) ((DOMSource) aggregatedFault
                                    .getContent()).getNode();
                            PatternHelper.aggregateContent(receivedMessage, aggregatedDocument,
                                    EIPConstants.FAULT_ELEMENT, SPLITTER_NAMESPACE);
                            aggregatedFault.setContent(UtilFactory.getSourceUtil()
                                    .createStreamSource(aggregatedDocument));
                            PatternHelper.copyAttachments(receivedMessage, aggregatedFault);
                            PatternHelper.copyProperties(receivedMessage, aggregatedFault);
                        } else {
                            asyncContext.setTerminated();
                            this.logger
                            .warning(this.getName()
                                    + ": terminate the pattern processing due to a Fault for the exchange with id '"
                                    + asyncContext.getOriginalExchange().getExchangeId()
                                    + "'");

                            asyncContext.getOriginalExchange().setFault((Fault) receivedMessage);
                            context.send(asyncContext.getOriginalExchange());
                        }
                    } else if (!asyncContext.isExceptionRobust()) {
                        if (exchange.getError() != null) {
                            asyncContext.setTerminated();
                            this.logger
                            .warning(this.getName()
                                    + ": terminate the pattern processing due to an error for the exchange with id '"
                                    + asyncContext.getOriginalExchange().getExchangeId()
                                    + "'");

                            asyncContext.getOriginalExchange().setError(exchange.getError());
                            context.send(asyncContext.getOriginalExchange());
                        }
                    }
                    if (asyncContext.decrementAndGetOngoingSubMessages() == 0
                            && !asyncContext.isTerminated()) {
                        this.logger.fine(this.getName() + ": terminate the pattern");
                        if (asyncContext.getOriginalExchange().getFault() == null) {
                            asyncContext.getOriginalExchange().setDoneStatus();
                        } else if (asyncContext.isFaultRobust()) {
                            final Document document = UtilFactory.getSourceUtil().createDocument(
                                    asyncContext.getOriginalExchange().getFault().getContent());
                            document.normalizeDocument();
                            asyncContext.getOriginalExchange().getFault().setContent(
                                    UtilFactory.getSourceUtil().createStreamSource(document));
                        }
                        context.send(asyncContext.getOriginalExchange());
                    }
                }
            } catch (final JBIException e) {
                if (!asyncContext.isExceptionRobust()
                        && asyncContext.getOriginalExchange().isActiveStatus()) {
                    asyncContext.getOriginalExchange().setError(new Exception(e));
                }
            }
        }
        return false;
    }

    /*
     * (non-Javadoc)
     * 
     * @see
     * org.ow2.petals.se.eip.patterns.AbstractPattern#validateMEP(java.net.URI)
     */
    @Override
    public boolean validateMEP(final URI mep) {
        return MEPConstants.IN_ONLY_PATTERN.equals(mep)
        || MEPConstants.ROBUST_IN_ONLY_PATTERN.equals(mep);
    }

    /*
     * (non-Javadoc)
     * 
     * @see org.ow2.petals.se.eip.patterns.Pattern#init()
     */
    public void init() {
    }
}
